# Vgrid 	1-D and 2-D interpolation program of velocity profiles
#		Use as an input the output of vpick
#		and the output can be used with Nmo 
#
#	Vgrid >Headout par=
#
#	par:
#	option=1	1 > 1-D interpoltion of nmid profiles. n2 must be
#				equal nmid as default.
#			2 > 2-D interpolation of nmid profiles with n2 output
#				profiles
#			3 > 2-D read a SEPLIB file, profiles interpolated in time
#			    suitable for output from taal:Overlay
#	vfile=vpick	input ascii file with the picked velocities (output of vpick),
#			except for option=3 vfile=file.H (seplib)
#	xmidpts=      Seplib file, midpoint numbers of input vrms profiles,
#			float format, esize=4
#	nmid=1		number of input profiles to interpolate.
#			cannot be bigger than the picked midpoints
#	n1=		number of time points in the interpolated profiles
#	d1=.004		sampling time of the profiles	
#	o1=0.		first time sample of the profiles
#	n2=nmid		number of output velocity profiles
#	niter1=0	number of conjugate gradient iteration for the 
#			interpolation in time 
#			(niter1=0 means linear interpolation)
#	niter2=0	number of conjugate gradient iteration for the 
#			interpolation in space 
#			(niter2=0 means linear interpolation)
#	alpha=0. 	coefficient for determining the slope for the 
#			linear extrapolation of the 
#			velocity profile after the last picked velocity.
#			slope=alpha*(average slope in the profile)
#			alpha=0 means flat extrapolation
#
#	BUG:		there is no o2; in output cmp numbers go from 1 to n2
#
#	ALGORITHM
#		The algorithm smoothes an initial linear interpolation
#		minimizing the second derivative with a constrained
#		conjugate gradient procedure.
#		The 2-D interpolation is done with 1-D interpolations
#		on the time axis followed by 1-D interpolations on the
#		space axis using the result of the time interpolations
#		as constraints.
#
#
#	Biondo Biondi 6/18/86
#%
# ----------
#  Keyword:  interpolation velocity-analysis grid : vpick
#  Keyword:  velocity-analysis interpolation
# ----------
#
#	Edit history
#	18/6/86	Biondo		author
#	10/12/86      Clement       added option=3
#
integer n1,n2,option,niter1,niter2,nmid
character*1  vfile(60)
from par:	integer option=1,niter1=0,niter2=0,nmid=1
from par:	integer n1,n2=nmid
from par:	real alpha=0.,d1:dt=.004,o1:t0=0.
from par:	character vfile
allocate: 	real b(n1),g(n1),s(n1),res(n1),gg(n1),ss(n1)
allocate: 	real b2(n2),g2(n2),s2(n2),res2(n2),gg2(n2),ss2(n2)
allocate:	real vel(n1,n2),hor(n2)
allocate:	integer npoint(nmid),pos(n1,nmid),mid(nmid)
 
subroutine vgrids(n1,n2,vfile,niter1,niter2,nmid,option,alpha,dt,t0,b,g,s,res,gg,ss,b2,g2,s2,res2,gg2,ss2,vel,hor,pos,npoint,mid)
integer infd,input,outfd,output,inf2,inf3,auxin
integer mode,ix,niter1,niter2,nmid,imid,option,midpos
integer n1,n2,i,j
integer npoint(nmid),pos(n1,nmid),mid(nmid)
character*60 vfile
real b(n1),g(n1),s(n1),res(n1),gg(n1),ss(n1)
real b2(n2),g2(n2),s2(n2),res2(n2),gg2(n2),ss2(n2)
real vel(n1,n2),hor(n2),dt,t0,alpha

outfd=output()
call hclose()


if((option!=1)&(option!=2)&(option!=3))call erexit('option must be equal 1,2or3')

if (option==1)	{		# 1-D interpolation
	if(n2!=nmid)call erexit('if option=1  n2 must be equal to nmid (as default)')
	#	read the ascii file 'vfile'
	call reasci(vfile,n1,n2,option,nmid,npoint,dt,t0,pos,mid,vel)
	#	interpolates nmid profiles
	do imid=1,nmid	{

		midpos=imid


		#	computation of the known term of the least square 
		#  	system	
		call aprod(0,n1,n1,vel(1,midpos),b,npoint(imid),
			pos(1,imid))

		#	linear interpolation
		call linin(n1,npoint(imid),vel(1,midpos),pos(1,imid),alpha)

		#	conjugate gradient smoothing
		call cg(n1,n1,vel(1,midpos),b,g,s,res,gg,ss,niter1,
			npoint(imid),pos(1,imid))

			}

		}
if (option==2)	{		#	2-D interpolation
	#	read the ascii file 'vfile'
	call reasci(vfile,n1,n2,option,nmid,npoint,dt,t0,pos,mid,vel)
	#	interpolates in time nmid profiles
	do imid=1,nmid	{

		midpos=mid(imid)

		#	computation of the known term of the least square 
		#  	system	
		call aprod(0,n1,n1,vel(1,midpos),b,npoint(imid),
			pos(1,imid))

		#	linear interpolation
		call linin(n1,npoint(imid),vel(1,midpos),pos(1,imid),alpha)

		#	conjugate gradient smoothing	
		call cg(n1,n1,vel(1,midpos),b,g,s,res,gg,ss,niter1,
			npoint(imid),pos(1,imid))
			}
				}
if (option==3)	{		#  no interpolation in time

	do j=1,n2 ; do i=1,n1 ; vel(i,j)=0.

	inf2=auxin('xmidpts')	# read midpoints
	call reed(inf2,hor,nmid*4)
	do j=1,nmid
	mid(j)= int( hor(j) )

	inf3=auxin('vfile')     # read vrms model
	do j=1,nmid
	call reed(inf3,vel(1,mid(j)),n1*4)

			}

if (option >= 2 )  {
	#	interpolates in space the velocity function
	#	at each time
	do i=1,n1	{
		#	the time interpolation is used as a contraint
		#	for the space interpolation
		do j=1,n2
			hor(j)=vel(i,j)



		#	computation of the known term of the least square 
		#  	system	
		call aprod(0,n2,n2,hor,b2,nmid,mid)

		#	linear interpolation
		call linin(n2,nmid,hor,mid,0)

		#	conjugate gradient smoothing
		call cg(n2,n2,hor,b2,g2,s2,res2,gg2,ss2,niter2,nmid,
			mid)

		do j=1,n2
			vel(i,j)=hor(j)
			}
		}
call srite('out',vel,n1*n2*4)
 		
stop;	end


#	subrotine that read an ascii file in the format of the output
#	of vpick program
subroutine reasci ( vfile , n1 ,n2,opt,nmid,npoint,dt,t0,pos,mid,vel ) 
integer n1,n2,nmid
integer cmid,p,opt
integer npoint(nmid),pos(n1,nmid),mid(nmid)
character*60 vfile
character*30 st
real vel(n1,n2),t,v,dt,t0


open(unit=7,file=vfile,status='old')

cmid=0

if(opt==1)	{
while (cmid<=nmid)	{
			{
	read(7,100,end=1000)st
	if(st(1:1)=='m')	{	#	if a new midpoint
		cmid=cmid+1
		npoint(cmid)=0
		decode(8,200,st(3:11))mid(cmid)	#position of the new midpoint
				}
	else			{
		decode(6,300,st(5:11))t		#picked time
		decode(9,400,st(13:21))v	#picked velocity
		if((t!=0.)|(v!=0.)) 	{	#if not an empty line
			npoint(cmid)=npoint(cmid)+1
			p=1+int((t-t0)/dt)
			if((p>n1)|(p<1))call erexit('a picked point is outside the interpolation interval (n1 and/or o1 are wrong')
			pos(npoint(cmid),cmid)=p
			vel(p,cmid)=v
					}
				}
			}
			}
		}
if (opt==2)	{
while ((cmid<=nmid)&(n2>=mid(cmid)))	{
			{
	read(7,100,end=1000)st
	if(st(1:1)=='m')	{	#if a new midpoint
		cmid=cmid+1
		npoint(cmid)=0
		decode(8,200,st(3:11))mid(cmid) #new midpoint position
				}
	else			{
		decode(6,300,st(5:11))t		#picked time
		decode(9,400,st(13:21))v	#picked velocity
		if((t!=0.)|(v!=0.)) 	{	#if not an empty line
			npoint(cmid)=npoint(cmid)+1
			p=1+int((t-t0)/dt)
			if((p>n1)|(p<1))call erexit('a picked point is outside the interpolation interval (n1 and/or o1 are wrong')
			pos(npoint(cmid),cmid)=p
			vel(p,mid(cmid))=v
					}
				}
			}
			}
		}
cmid=cmid-1
1000 continue


if(cmid<nmid)	call erexit(' The midpoints number is too high')

100 format(a)
200 format(i8)
300 format(f6.4)
400 format(f9.4)

return;	end


#	subroutine to interpolate linearly uneven sampled data

	subroutine linin(n1,lpos,data,pos,alpha)
	
	integer n1,lpos,pos(lpos),id,ip,left,right
	real data(n1),rl,rr,slope,alpha

	# 	 zero slope boundary condition
	do id=1,pos(1)	
		data(id)=data(pos(1))

	do ip=1,lpos-1	{
		left=pos(ip)
		right=pos(ip+1)
		rl=data(left)
		rr=data(right)
		slope=(rr-rl)/(right-left)
		do id=left,right
			data(id)=rl+slope*(id-left)
			}
	#	alpha*(average slope) boundary condition
	slope=alpha*(data(pos(lpos))-data(pos(1)))/(pos(lpos)-pos(1))
	do id=pos(lpos),n1
		data(id)=rr+slope*(id-right)

	return;	end



#	conjugate gradient subroutine following Jon's
#	Conjugate Gradient Tutorial SEP-48
subroutine cg(n,m,x,b,g,s,res,gg,ss,niter,leniw,iw)

integer n,m,iter,niter,i,j,leniw
integer iw(leniw)
real x(n),b(m),res(m),g(n),s(n),gg(m),ss(m)
real dot,sds,gdg,gds,determ,gdr,sdr,alfa,beta,rem

if (niter!=0)	{
do i=1,n		# save initial values
	s(i)=x(i)
do i=1,m
	res(i)=0.
call aprod(1,m,n,x,res,leniw,iw)
do i=1,n		# restore initial value
	x(i)=s(i)
do i=1,m
	res(i)=b(i)-res(i)
do iter=0,niter	{

	do i=1,n
		g(i)=0.
	call aprod(2,m,n,g,res,leniw,iw)

	do i=1,m
		gg(i)=0.
	call aprod(1,m,n,g,gg,leniw,iw)

	if (iter==0)	{
		alfa=dot(m,gg,res)/dot(m,gg,gg)
		beta=0.
			}
	else	{
		gdg=dot(m,gg,gg)
		sds=dot(m,ss,ss)
		gds=dot(m,gg,ss)
		determ= gdg*sds - gds*gds +1.e-30
		gdr=dot(m,gg,res)
		sdr=dot(m,ss,res)
		alfa=(sds*gdr - gds*sdr)/determ
		beta=(-gds*gdr + gdg*sdr)/determ
		}
	do i=1,n
		s(i)=alfa*g(i) + beta*s(i)
	do i=1,m
		ss(i)=alfa*gg(i) + beta*ss(i)
	do i=1,n
		x(i)=x(i) + s(i)
	do i=1,m
		res(i)=res(i) - ss(i)	
		}
		}
return;	end


real function dot(n,x,y)
integer i,n
real val,x(n),y(n)
val=0.
do i=1,n
	val=val + x(i)*y(i)
dot=val
return;	end



#	subroutine that computes the second derivatives of a vector
#	suitable to be used in a conjugate gradient program as LSQR
#	with contrained variables indicated by the content of the
#	vector iw
	subroutine aprod(mode,m,n,x,y,leniw,iw)

	integer m,n,leniw,iw(leniw),mode,iy,ix,ic
	real x(n),y(m)

	if (mode==0)	{	# calculation of the known term 
		do iy=2,m-1	# N.B. change of sign
			y(iy)=-(x(iy-1)-2.*x(iy)+x(iy+1))
		y(1)=-(-1*x(1)+x(2))
		y(m)=-(x(m-1)-1*x(m))
			}

	#	y = y + (second derivatives of x)
	if (mode==1)	{
		do ic=1,leniw	#contrained variables set to zero
			x(iw(ic))=0.
		do iy=2,m-1
			y(iy)=y(iy)+x(iy-1)-2.*x(iy)+x(iy+1)
		y(1)=y(1)-1*x(1)+x(2)	# end effects
		y(m)=y(m)+x(m-1)-1*x(m)	# end effects
			}

	#	x = x + (transpose of second derivatives of y)
	if (mode==2)	{
		do ix=2,n-1
			x(ix)=x(ix)+y(ix-1)-2.*y(ix)+y(ix+1)
		x(1)=x(1)-1*y(1)+y(2)	# end effects
		x(n)=x(n)+y(n-1)-1*y(n)	# end effects
		do ic=1,leniw		#contrained variables set to zero
			x(iw(ic))=0.
			}
	return
	end


