# The ultimate modeling machine !
#
# General explicit Finite Differences
# 
# Modeling and RTM
#
# Ultimod par=Par <Hsourcefn >junk
#
#
#
# type = 1     ACOUSTIC
#              ELASTIC:
# type = 2     ISOTROPIC
# type = 3     CUBIC
# type = 4     TI
# type = 5     ORTHO 
# type = 6     MONO 
# type = 7     TRICLINIC 
#
#
#  moduli = header of velocity model 
#
#    ACOUSTIC      (K,rho)
#
#    ISOTROPIC     (c11,c55,rho)
#    CUBIC         (c11,c13,c55,rho)
#
#    TRANSISO 2d2c (c11,c13,c33,c55,rho)
#    TRANSISO   3c (c11,c13,c33,c55,C66,rho)
#
#    ORTHO    2d2c (c11,c13,c33,c55,rho)
#    ORTHO    2d3c (c11,c13,c33,c55,c44,c66,rho)
#    ORTHO    3d3c (c11,c13,c33,c55,c44,C66,c12,c22,c23,rho)
#
#    MONO          (c11,c12,c13,c15,c22,c23,c25,c33,c35,c44,c46,c55,C66,rho)
#
#    TRICLINIC     (c11,c12,c13,c14,c15,c16,c22,c23,c24,c25,c26,c33,c34,c35,c36,
#                   c44,c45,c46,c55,c56,C66,rho)
#
#    from moduli:
#        n1 = # of z grid points 
#        n2 = # of x grid points 
#        n3 = # of y grid points 
#        d1 = z grid point spacing 
#        d2 = x grid point spacing 
#        d3 = y grid point spacing 
#
#
#    from stdin:
#   Hsource = Source function header (stdin)
#        n1 = length of source
#        d1 = time sample interval
#             so it should be within stability limits
#
#    UNITS:   stiffnesses are in [10^9   N/m^2]   (typical numbers 0.-100.)
#             density     is  in [103^3 kg/m^3]   (typically  1.-5.)
#             time        is  in [sec]            (typically  0.004)
#             distances   are in [km]             (typically  0.025)
#
#
#   From the command line or parameter file:
#
#   modeling   = 1  do modeling otherwise reverse time migration
#   nt         =    number of time steps to run
#   nsrc       = 1  # of shots to make (default=1)
#   src_1      =    Position of first source (in grid points)
#
#   src_type = 0    acceleration force 
#                   (multicomponent) Hsourcefn has n2=2 (z,x) or n2=3 (z,x,y)
#                   eg. (0,0,1) source vector would usually be pure sh waves
#
#   src_type = 1    pressure source  (give three identical components n3=3)
#                                    (normal stresses)  (buggy, try just one)
#   src_type = 2    shear soure (tangential stresses) a (1,0,0) vector would
#                               give (z,x) in plane waves only
#
#   src_type = 3    shear source (give three components, the first two are
#                   generating a couple force( z-x), the last is y displacement
#
#  src_inc =        shot increment (in x grid points) default=1
#  group_1 =        Position of first receiver for first shot default=1
#  src_depth =      Shot depth (grid points)
#
#  surf_type =      free surface (default) =1; absorbing =0
#  snap_i    =      number of time steps between snapshots
#  wantseis  = 1    write out seismograms
#  show      = 1    bring up the Xwindow display
#  ishow     = 1    increment of time steps to be shown
#
#  Stability:
#                   for stability vpmax*dt/min(dx,dz) <.5
#                   for aniso: approx. sqrt(max(c11,c33)/rho)*dt/min(dx,dz) <.5
#
#                   for no dispersion (vmin/fmax)/max(dx,dz) >2.8
#% end of self doc
# keywords: anisotropy 2d 3D elastic acoustic modeling RTM
#

#include "commons"

	integer nt,ntmax
        real dz,dx,dy

#	integer infd1,outfd1,infd2,outfd2,outfd3,outfd4,outfd5,outfd6,imagfd
	integer auxin,auxout
	integer nsrc,src_1,src_inc,src_depth,izmin
	integer ngroup,group_1,rollalong,snap_i 
	real 	vpmax,stab
	integer modeling,ntmod,nmodpar,wantseis
	integer show,ishow
	integer centered
	

	# what type of medium:
###        call testdn()
	from par:      integer type:type
	from par:      integer ncomp:ncomp=2
	from par:      integer wantseis:wantseis=1

	# modeling or RTM
	from par:      integer modeling:modeling=1

	# the elastic moduli are in a file par=moduli
	#
	# the moduli file has to have nx and nz the size of the model.
        # for TI 2comp
	# the order they are in is c11,c13,c33,c44,rho
	#
        # for TI 3comp
	# the order they are in is c11,c13,c33,c44,C66,rho
	#

#       the standard model input:

        from aux:       moduli  integer n1:nz , n2:nx, n3:ny=1, n4:nmodpar=0
        from aux:       moduli  real    d1:dz , d2:dx, d3:dy=0.


	if (nmodpar==0 | nmodpar==1 ) {	         #we have a degenerate 2D model
        from aux:       moduli  integer n3:nmodpar
	ndim = 2		# it is only a 2D problem
	ny = 1 ; dy = 0.
	}
	else {
	ndim = 3		# it is a 3D problem
	}

	
	if ( nmodpar != 5 & ndim == 2 & ncomp == 2 & type == TRANSISO ) {
	write (0,*) 'ndim',ndim,' ncomp',ncomp,' needs',4,_
                    ' stiffnesses you got only',nmodpar-1
	call exit(-1)
	}
	if ( nmodpar != 6 & ndim == 2 & ncomp == 3 & type == TRANSISO ) {
	write (0,*) 'ndim',ndim,' ncomp',ncomp,' needs',5,_
                    ' stiffnesses you got only',nmodpar-1
	call exit(-1)
	}

	# get sample interval of sourcefunction
        from history:   integer n1:nts, n2:nscomp
        from history:   real    d1:dt
	if (ntcomp>3) { 
        write(0,*)'number of source components >3' ;call exit(-1) }
	# the number of time steps
	from par:	integer nt:nt=10
        from par:       integer ngroup:ngroup=nx

	from par:       integer ntmod:ntmod=nt



ntmax=max(nt,nts)
ndamp=20
nabsorb=4
nbound = ndamp+nabsorb

# do some intialization
nbounds(1)=ndamp+nabsorb; nbounds(2)=ndamp+nabsorb; nbounds(3)=0
nbound1=ndamp+nabsorb ; nbound2=ndamp+nabsorb; nbound3=0;

nloc=1          #number of sources



        allocate:  real uxout((nz+2*35)*(nx+2*24)),uzout((nz+2*35)*(nx+2*24))

	if (ncomp == 3) {
        allocate:  real uyout((nz+2*35)*(nx+2*24))
	}

	allocate:  real modelparam((nz+2*35),(nx+2*24),(ny+2*nbound3),nmodpar)

        allocate:  real sourcef(ntmax,3)

# should allocate only nx*nz and write out for large nt

####ifdef DONT_KEEP_SEIS_ONFE
#        ntseis = nz
####else
        ntseis = ((nt-1)/nz+1)*nz
####endif

    if (wantseis==1) {

        allocate:  real uzseis(nx,ntseis), uxseis(nx,ntseis)
        allocate:  real vzseis(nx,ntseis), vxseis(nx,ntseis)
#        allocate:  real uzseis(nx,2*ntseis), uxseis(nx,2*ntseis)
#        allocate:  real vzseis(nx,2*ntseis), vxseis(nx,2*ntseis)
        if ( ncomp == 3 ) {
        allocate:  real uyseis(nx,ntseis), vyseis(nx,ntseis)
#        allocate:  real uyseis(nx,2*ntseis), vyseis(nx,2*ntseis)
	}
    }else{
        # dummy allocation
        allocate:  real uzseis(1,1), uxseis(1,1)
        allocate:  real vzseis(1,1), vxseis(1,1)
        if ( ncomp == 3 ) {
        allocate:  real uyseis(1,1), vyseis(1,1)
        }
    }

        allocate:  real mult(35)


	# the sourcefunction is the redirected standard input
	# only one input sourcefile

	# 
	# Now fetch a set of parameters from the command line

	# the number of time steps
	#from par:	integer nt:nt=10
	# get the number of sources default=1
	from par:	integer nsrc:nsrc=1
	# the first shot location (in x grid points)
	from par:       integer src_1:src_1=nx/2
	# the shot location increment (in x grid points)
	from par:       integer src_inc:src_inc=1
	# the shot depth (in z grid points) default at surface
	from par:       integer src_depth:src_depth=1
	# sourcetype	default vertical
	from par:       integer src_type:src_type=1
	# surface boundary condition type  default free surface
	from par:       integer surf_type:surf_type=1
      	if(surf_type==1){
        	izmin=1 }
      	else {
        	izmin=-(ndamp-1)
      	}
	# the location of the first trace (in x gridpoints)
	from par:       integer group_1:group_1=1
	# the number of traces to record default all surface points
	from par:       integer ngroup:ngroup=nx
	# geophone depth
	from par:       integer geo_depth:geo_depth=1
	# roll the receiver spread along with the shots (const geometry recordg)
	from par:       integer rollalong:rollalong=1
	# snapshot increment default no snapshots
	from par:       integer snap_i:snap_i=nt+1
	from par:       integer centered:centered=0
      	from either:    integer oneway:oneway=1	
      	from par:       integer show:show=0	
      	from par:       integer timing:timing=0	
	from par:       integer ishow:ishow=1
	from par:       real damping=0.002
	

	# now putch out info for the surface vertical geophones

	if (modeling == 1) {

      		call auxputch('n2','i',nt,'zseis')
      		call auxputch('n1','i',ngroup,'zseis')
      		call auxputch('n3','i',nsrc,'zseis')
      		call auxputch('rollalong','i',rollalong,'zseis')
      		call auxputch('surf_type','i',surf_type,'zseis')
      		call auxputch('geo_depth','i',geo_depth,'zseis')
      		call auxputch('src_1','i',src_1,'zseis')
      		call auxputch('src_depth','i',src_depth,'zseis')
      		call auxputch('src_inc','i',src_inc,'zseis')
      		call auxputch('d2','r',dt,'zseis')
      		call auxputch('d1','r',dx,'zseis')
      		call auxputch('group_1','i',group_1,'zseis')
      		call auxputch('o1','r',(group_1-1)*dx,'zseis')

	# now putch out info for the surface inline horizontal geophones
	
        	call auxputch('n2','i',nt,'xseis')
        	call auxputch('n1','i',ngroup,'xseis')
        	call auxputch('n3','i',nsrc,'xseis')
        	call auxputch('rollalong','i',rollalong,'xseis')
        	call auxputch('surf_type','i',surf_type,'xseis')
        	call auxputch('geo_depth','i',geo_depth,'xseis')
        	call auxputch('src_1','i',src_1,'xseis')
        	call auxputch('src_depth','i',src_depth,'xseis')
        	call auxputch('src_inc','i',src_inc,'xseis')
        	call auxputch('d2','r',dt,'xseis')
        	call auxputch('d1','r',dx,'xseis')
        	call auxputch('group_1','i',group_1,'xseis')
        	call auxputch('o1','r',(group_1-1)*dx,'xseis')

	# cross line horizontal geophone

	        if ( ncomp == 3 )  {
        	call auxputch('n2','i',nt,'yseis')
        	call auxputch('n1','i',ngroup,'yseis')
        	call auxputch('n3','i',nsrc,'yseis')
        	call auxputch('rollalong','i',rollalong,'yseis')
        	call auxputch('surf_type','i',surf_type,'yseis')
        	call auxputch('geo_depth','i',geo_depth,'yseis')
        	call auxputch('src_1','i',src_1,'yseis')
        	call auxputch('src_depth','i',src_depth,'yseis')
        	call auxputch('src_inc','i',src_inc,'yseis')
        	call auxputch('d2','r',dt,'yseis')
        	call auxputch('d1','r',dx,'yseis')
        	call auxputch('group_1','i',group_1,'yseis')
        	call auxputch('o1','r',(group_1-1)*dx,'yseis')
	        }
      	}	
	else{
# use zseis to get some information

      		from aux:      zseis integer n2:nt
      		from aux:      zseis integer n1:ngroup
		from aux:      zseis integer n3:nsrc
		from aux:      zseis real d1:dx
		from aux:      zseis real d2:dt
		from aux:      zseis integer rollalong:rollalong
		from aux:      zseis integer surf_type:surf_type
		from aux:      zseis integer geo_depth:geo_depth
		from aux:      zseis integer src_1:src_1
		from aux:      zseis integer src_depth:src_depth
		from aux:      zseis integer src_inc:src_inc

        	to aux:        image integer n1:nz
        	to aux:        image integer n2:nx
        	to aux:        image integer n3:4
        	to aux:        image integer n4:1
        	to aux:        image real d1:dz
        	to aux:        image real d2:dx

	if ( 1<=snap_i & snap_i<=nt-1) {

        	to aux:        xrevshot integer n1:nz
        	to aux:        xrevshot integer n2:nx
        	to aux:        xrevshot integer n3:(nt-1)/snap_i*nsrc
        	to aux:        xrevshot integer snap_i:snap_i*dt
        	to aux:        xrevshot real d1:dz
        	to aux:        xrevshot real d2:dx

        	to aux:        zrevshot integer n1:nz
        	to aux:        zrevshot integer n2:nx
        	to aux:        zrevshot integer n3:(nt-1)/snap_i*nsrc
        	to aux:        zrevshot integer snap_i:snap_i*dt
        	to aux:        zrevshot real d1:dz
        	to aux:        zrevshot real d2:dx

		if ( ncomp == 3 ) {
        	to aux:        yrevshot integer n1:nz
        	to aux:        zrevshot integer n2:nx
        	to aux:        yrevshot integer n3:(nt-1)/snap_i*nsrc
        	to aux:        yrevshot integer snap_i:snap_i*dt
        	to aux:        yrevshot real d1:dz
        	to aux:        yrevshot real d2:dx
                }	


        	to aux:        imsnap integer n1:nz
        	to aux:        imsnap integer n2:nx
        	to aux:        imsnap integer n3:4
        	to aux:        imsnap integer snap_i:snap_i*dt
        	to aux:        imsnap real d1:dz
        	to aux:        imsnap real d2:dx
        	to aux:        imsnap integer n4:(nt-1)/snap_i*nsrc
	 }

	}
	
	# Putch out info for the snapshots of the wavefield (horizontal accel)
	
      	if(1<=snap_i & snap_i<=nt){
        	call auxputch('n1','i',nz,'xsnap')
        	call auxputch('n2','i',nx,'xsnap')
        	call auxputch('n3','i',nt/snap_i*nsrc,'xsnap')
        	call auxputch('snap_i','r',snap_i*dt,'xsnap')
        	call auxputch('d1','r',dz,'xsnap')
        	call auxputch('d2','r',dx,'xsnap')
	
	# Putch out info for the snapshots of the wavefield (vertical accel.)
	
        	call auxputch('n1','i',nz,'zsnap')
        	call auxputch('n2','i',nx,'zsnap')
        	call auxputch('n3','i',nt/snap_i*nsrc,'zsnap')
        	call auxputch('snap_i','r',snap_i*dt,'zsnap')
        	call auxputch('d1','r',dz,'zsnap')
        	call auxputch('d2','r',dx,'zsnap')

		if (ncomp == 3) {
        	call auxputch('n1','i',nz,'ysnap')
        	call auxputch('n2','i',nx,'ysnap')
        	call auxputch('n3','i',nt/snap_i*nsrc,'ysnap')
        	call auxputch('snap_i','r',snap_i*dt,'ysnap')
        	call auxputch('d1','r',dz,'ysnap')
        	call auxputch('d2','r',dx,'ysnap')
		}
      	}
	
	
	#
	# the elastic moduli are in a file par=moduli
	#
	
	# close up the headers and get to work
	
      	call hclose()

	
	subroutine ti2d(modeling,uxout,uzout,_         
	nmodpar,modelparam,_
	uxseis,uzseis,uyseis,vxseis,vzseis,vyseis,ntseis,sourcef,mult,nt,_
        ntmax,nbound,dz,dx,_
	nsrc,src_1,src_inc,src_depth,izmin,_
	ngroup,group_1,rollalong,snap_i,_
        show,ishow,centered,timing,ntmod,wantseis)

	#infd1,outfd1,infd2,outfd2,outfd3,outfd4,outfd5,outfd6,imagfd,imsnapfd,_

#include "commons"


        integer nt,ntmax,nbound,modeling,ntseis,iparam,k
        real dz,dx,dy
	real stab

#        integer infd1,outfd1,infd2,outfd2,outfd3,outfd4,outfd5,outfd6
#	integer imagfd,imsnapfd
        integer input,output
        integer nsrc,src_1,src_inc,src_depth,izmin
        integer ngroup,group_1,rollalong,snap_i,show
	integer centered,timing,ntmod,ishow,wantseis
	integer nmodpar

      real uxout(nz+2*nbound,nx+2*nbound),uzout(nz+2*nbound,nx+2*nbound)

real modelparam(nz+2*nbound1,nx+2*nbound2,ny+2*nbound3,nmodpar)
#mf$layout modelparam(:serial,:serial,:serial,:serial)

      real uzseis(nx,ntseis),uxseis(nx,ntseis),uyseis(nx,ntseis)
      real vzseis(nx,ntseis),vxseis(nx,ntseis),vyseis(nx,ntseis)
      real mult(-34:0),sourcef(nts,3)
#      real mult(-34:0),sourcef(ntmax,3)

      integer i,j

        write(0,*)' You have entered the TWILIGHT ZONE'
	write(0,*)'ntmod',ntmod

      	call sreed('in',sourcef,4*nts*nscomp)
	
	# now read in the various elastic moduli and the density from
	# one file the order is c11 c13 c33 c44 rho


 if (type == ACOUSTIC ) { 
      write(0,*)' Did I read the correct number of parameters: acoust'}
	
	
 do iparam=1,nmodpar {
   do  k=1,ny
   do  j=1,nx
   call sreed('moduli',modelparam(1+nbound1,j+nbound2,k+nbound3,iparam),4*nz)
 }

## stability analysis
##	stab=0
##	do i=1+nbound,nz+nbound
##	 do j=1+nbound,nx+nbound
##	   stab = max(stab,sqrt(max(c11(i,j),c33(i,j))/rho(i,j))*dt/min(dx,dz))
##        if(stab<=0.5){
##		write(0,*)'stability from stiffnesses',stab
##                call putch('vpmax*dt/(min(dz,dx))','r',stab)
##		stab=1.e6
##		do i=1+nbound,nz+nbound
##	 		do j=1+nbound,nx+nbound
##	   		  stab =min(stab,sqrt(min(c11(i,j),c33(i,j))/rho(i,j)))
##		stab = (stab*2.*dt)/max(dx,dz)
##		write(0,*)'dispersion at Nyquist',stab,'   should be >2.8'
##            	call putch('dispersion','r',stab) 
##		}
##	        else {
##                write(0,*)'stability not ok check your stiffnesses = ',stab
##                stop
##        }
## end of stability

if (type == ACOUSTIC ) {
        call initcommon(ndim,1,1./dx,0.,1./dz,9)
}
else {
        call initcommon(ndim,ncomp,1./dx,0.,1./dz,9)
} 
	
	call CMF_init()

        call maunaloa(modeling,uxout,uzout,uxseis,uzseis,uyseis,_
        vxseis,vzseis,vyseis,ntseis,_
        nmodpar,modelparam,_
        sourcef,mult,nt,ntmax,nbound,dz,dx,_
        nsrc,src_1,src_inc,src_depth,izmin,_
        ngroup,group_1,rollalong,snap_i,_
        show,ishow,timing,centered,ntmod,wantseis)

      return
      end
